#lang racket
(require rackunit)

(define (r:value? v)
  (or (r:number? v)
      (r:void? v)
      (r:lambda? v)))
(struct r:void () #:transparent)
(struct r:number (value) #:transparent)
(struct r:lambda (params body) #:transparent)
(define (r:expression? e)
  (or (r:value? e)
      (r:variable? e)
      (r:apply? e)))
(struct r:variable (name) #:transparent)
(struct r:apply (func args) #:transparent)

; 10
(r:number 10) ; (r:number 10)
(r:number-value (r:number 10)) ; 10
;; (r:void) ; (r:void)
; (r:number-value 10) -> error
;; (r:number "foo") ; (r:number "foo")
; (r:number-value "foo") ; -> error
; (lambda () 10)
;; (r:lambda (list) (list (r:number 10)))
; (r:lambda (r:number 10))
; (r:lambda (list) (r:number 10))
;; '(define f
;;    (lambda (y)
;;      (+ y 10)))

(define (cons x y)
  (lambda (b)  ; <- we use a parameter to choose which stored data to return
    (cond [b x] [else y]))) ; <- passing #t returns x
                            ;    passing #f returns y
; We now define our own 'car' and 'cdr'
(define (car f) (f #t))  ; Returns the first element of the pair
(define (cdr f) (f #f))  ; Returns the second element of the pair
(define p (cons 10 20))  ; Same as: (define (p b) (cond [b 10] [else 20])))
;; (car p)                  ; Returns 10 because (car p) -> (p #f) -> 10
;; (cdr p)                  ; Returns 20 because (cdr p) -> (p #t) -> 20

(struct frozen (func arg1) #:transparent)

(define f (frozen + 10))
;; (frozen-func f)
;; (frozen-arg1 f)

(define (call f arg2)
  (define func (frozen-func f))
  (define arg1 (frozen-arg1 f))
  (func arg1 arg2))

;; (call f 30)

(define (pipeline l n)
  (cond [(empty? l) n]
        [else
          (define f (first l))
          (define r (rest l))
          (pipeline r (f n))]))
;; (pipeline (list (lambda (x) (+ x 1)) (lambda (x) (* x 2))) 10)

(define (+-frozen x)
  (lambda (y)
    (+ x y)))

(define plus10 (+-frozen 10))
 ; (lambda (y) (+ 10 y))
;; (plus10 20)

(define (member element elements)
  (cond [(empty? elements) #f]
        [else
          (define f (first elements))
          (define r (rest elements))
          (or
            (equal? element f)
            (member element r))]))

(check-true (member 1 (list 3 6 1)))
(check-true (member #t (list 3 #t (list))))
(check-false (member 1 (list 3 #t (list 1))))
(check-false (member #f (list)))

;; (member 1 (list 3 5 6))
;; (member 1 (list 3 1 10))

(define (match-prefix? l prefix)
  (cond [(empty? l) #f]
        [else
          (define f (first l))
          (define r (rest l))
          (or (string-prefix? f prefix)
              (match-prefix? r prefix))]))

(check-true (string-prefix? "Racket" "R") ) ; available in standard library
(check-true (match-prefix? (list "foo" "Racket") "R"))
(check-false (match-prefix? (list "foo" "bar") "R"))
(check-false (match-prefix? (list) "R"))

;(list + *)

;; (define (apply1 fr arg)
;;   (define func (frozen-func fr))    ; Bind a function to a local variable
;;   (define arg1 (frozen-arg1 fr))
;;   (func arg1 arg))                  ; Call a function bound to a local variable




#|
; (lambda (x) 10 x)
'(lambda (x) 10 x)
(r:lambda (list (r:variable 'x)) (list (r:number 10) (r:variable 'x)))|#

